-- Paraselene's WildRace
-- Version 0.5.2

require "Apollo"

local CHARACTER_DATA_FORMAT_VERSION = 6
local WINDOW_HORIZONTAL_GAP = 3

ParaseleneWildRace.CharacterData = {}

function ParaseleneWildRace.CharacterData:new(o)
    if type(o) ~= "table" then
        return nil
    end

    local i = {}
    setmetatable(i, self)
    self.__index = self 

    i.version = ((type(o.version) == "number") and o.version) or nil
    local restoreSucceeded = false
    if (i.version == 1) or (i.version == 2) then
        restoreSucceeded = i:RestoreV1V2(o)
    elseif i.version == 3 then
        restoreSucceeded = i:RestoreV3(o)
    elseif i.version == 4 then
        restoreSucceeded = i:RestoreV4(o)
    elseif i.version == 5 then
        restoreSucceeded = i:RestoreV5(o)
    elseif i.version == 6 then
        restoreSucceeded = i:RestoreV6(o)
    end
    i.version = CHARACTER_DATA_FORMAT_VERSION

    return (restoreSucceeded and i) or nil
end

function ParaseleneWildRace.CharacterData.CreateDefault()
    local clientWindowInfo = ParaseleneWildRace.CharacterData.CreateDefaultClientWindowInfo()
    local editorWindowInfo = ParaseleneWildRace.CharacterData.CreateDefaultEditorWindowInfo(clientWindowInfo)
    local filesWindowInfo = ParaseleneWildRace.CharacterData.CreateDefaultFilesWindowInfo(clientWindowInfo)
    local optionsWindowInfo = ParaseleneWildRace.CharacterData.CreateDefaultOptionsWindowInfo(clientWindowInfo)
    local shareWindowInfo = ParaseleneWildRace.CharacterData.CreateDefaultShareWindowInfo(clientWindowInfo)
    local trackMasterOptions = ParaseleneWildRace.CharacterData.CreateDefaultTrackMasterOptions()
    local uiOptions = ParaseleneWildRace.CharacterData.CreateDefaultUIOptions()
    local worldMarkerOptions = ParaseleneWildRace.CharacterData.CreateDefaultWorldMarkerOptions()
    local areWaypointsVisible = ParaseleneWildRace.CharacterData.CreateDefaultAreWaypointsVisible()
    local currentCourse = ParaseleneWildRace.CharacterData.CreateDefaultCurrentCourse()

    return ParaseleneWildRace.CharacterData:new{
        version = CHARACTER_DATA_FORMAT_VERSION,
        clientWindowInfo = clientWindowInfo,
        editorWindowInfo = editorWindowInfo,
        filesWindowInfo = filesWindowInfo,
        optionsWindowInfo = optionsWindowInfo,
        shareWindowInfo = shareWindowInfo,
        trackMasterOptions = trackMasterOptions,
        uiOptions = uiOptions,
        worldMarkerOptions = worldMarkerOptions,
        areWaypointsVisible = areWaypointsVisible,
        currentCourse = currentCourse
    }
end

function ParaseleneWildRace.CharacterData.CreateDefaultClientWindowInfo()
    local display = Apollo.GetDisplaySize()
    local windowGroupWidth = ParaseleneWildRace.minClientWindowWidth +
                             WINDOW_HORIZONTAL_GAP +
                             ParaseleneWildRace.minFilesWindowWidth -- Widest satallite window width.
    local windowGroupHeight = ParaseleneWildRace.minClientWindowHeight -- Tallest satallite window width.

    return ParaseleneWildRace.WindowInfo:new{
        left = (display and math.floor((display.nWidth - windowGroupWidth) / 2)) or 0,
        top = (display and math.floor((display.nHeight - windowGroupHeight) / 2)) or 0,
        width = ParaseleneWildRace.minClientWindowWidth,
        height = ParaseleneWildRace.minClientWindowHeight
    }
end

function ParaseleneWildRace.CharacterData.CreateDefaultEditorWindowInfo(clientWindowInfo)
    return ParaseleneWildRace.WindowInfo:new{
        left = clientWindowInfo.left + clientWindowInfo.width + WINDOW_HORIZONTAL_GAP,
        top = clientWindowInfo.top
    }
end

function ParaseleneWildRace.CharacterData.CreateDefaultFilesWindowInfo(clientWindowInfo)
    return ParaseleneWildRace.WindowInfo:new{
        left = clientWindowInfo.left + clientWindowInfo.width + WINDOW_HORIZONTAL_GAP,
        top = clientWindowInfo.top,
        width = ParaseleneWildRace.minFilesWindowWidth,
        height = ParaseleneWildRace.minFilesWindowHeight
    }
end

function ParaseleneWildRace.CharacterData.CreateDefaultOptionsWindowInfo(clientWindowInfo)
    return ParaseleneWildRace.WindowInfo:new{
        left = clientWindowInfo.left + clientWindowInfo.width + WINDOW_HORIZONTAL_GAP,
        top = clientWindowInfo.top
    }
end

function ParaseleneWildRace.CharacterData.CreateDefaultShareWindowInfo(clientWindowInfo)
    return ParaseleneWildRace.WindowInfo:new{
        left = clientWindowInfo.left + clientWindowInfo.width + WINDOW_HORIZONTAL_GAP,
        top = clientWindowInfo.top
    }
end

function ParaseleneWildRace.CharacterData.CreateDefaultTrackMasterOptions()
    return ParaseleneWildRace.TrackMasterOptions:new{
        enabled = false,
        lineNumber = 1
    }
end

function ParaseleneWildRace.CharacterData.CreateDefaultUIOptions()
    return ParaseleneWildRace.UIOptions:new{
        background = ParaseleneWildRace.backgrounds.holo
    }
end

function ParaseleneWildRace.CharacterData.CreateDefaultWorldMarkerOptions()
    return ParaseleneWildRace.WorldMarkerOptions:new{}
end

function ParaseleneWildRace.CharacterData.CreateDefaultAreWaypointsVisible()
    return false
end

function ParaseleneWildRace.CharacterData.CreateDefaultCurrentCourse()
    return ParaseleneWildRace.Course:new{
        name = "New Course",
        strictMode = false,
        timedMode = false,
        timeout = 0
    }
end

function ParaseleneWildRace.CharacterData:IsValid()
    -- self.version is not checked here.
    return (self.clientWindowInfo ~= nil) and
           (self.editorWindowInfo ~= nil) and
           (self.filesWindowInfo ~= nil) and
           (self.optionsWindowInfo ~= nil) and
           (self.shareWindowInfo ~= nil) and
           (self.trackMasterOptions ~= nil) and
           (self.uiOptions ~= nil) and
           (self.worldMarkerOptions ~= nil) and
           (self.areWaypointsVisible ~= nil) and
           (self.currentCourse ~= nil)
end

function ParaseleneWildRace.CharacterData:RestoreV1V2(o)
    -- self.version is not restored here.
    -- Note: V1 and V2 window positions are not preserved.
    self.clientWindowInfo = self.CreateDefaultClientWindowInfo()
    self.editorWindowInfo = self.CreateDefaultEditorWindowInfo(self.clientWindowInfo)
    self.filesWindowInfo = self.CreateDefaultFilesWindowInfo(self.clientWindowInfo)
    self.optionsWindowInfo = self.CreateDefaultOptionsWindowInfo(self.clientWindowInfo)
    self.shareWindowInfo = self.CreateDefaultShareWindowInfo(self.clientWindowInfo)
    self.trackMasterOptions = self.CreateDefaultTrackMasterOptions()
    self.uiOptions = self.CreateDefaultUIOptions()
    self.worldMarkerOptions = self.CreateDefaultWorldMarkerOptions()
    self.areWaypointsVisible = self.CreateDefaultAreWaypointsVisible()
    self.currentCourse = self.CreateDefaultCurrentCourse()

    return self:IsValid()
end

function ParaseleneWildRace.CharacterData:RestoreV3(o)
    -- self.version is not restored here.
    self.clientWindowInfo = ParaseleneWildRace.WindowInfo:new(o.clientWindowInfo)
    self.editorWindowInfo = ParaseleneWildRace.WindowInfo:new(o.editorWindowInfo)
    self.filesWindowInfo = ParaseleneWildRace.WindowInfo:new(o.filesWindowInfo)

    local clientWindowInfo = self.CreateDefaultClientWindowInfo()
    self.optionsWindowInfo = self.CreateDefaultOptionsWindowInfo(clientWindowInfo)
    self.shareWindowInfo = self.CreateDefaultShareWindowInfo(clientWindowInfo)
    self.trackMasterOptions = self.CreateDefaultTrackMasterOptions()
    self.uiOptions = self.CreateDefaultUIOptions()
    self.worldMarkerOptions = self.CreateDefaultWorldMarkerOptions()
    self.areWaypointsVisible = self.CreateDefaultAreWaypointsVisible()
    self.currentCourse = self.CreateDefaultCurrentCourse()

    return self:IsValid()
end

function ParaseleneWildRace.CharacterData:RestoreV4(o)
    -- self.version is not restored here.
    self.clientWindowInfo = ParaseleneWildRace.WindowInfo:new(o.clientWindowInfo)
    self.editorWindowInfo = ParaseleneWildRace.WindowInfo:new(o.editorWindowInfo)
    self.filesWindowInfo = ParaseleneWildRace.WindowInfo:new(o.filesWindowInfo)
    self.optionsWindowInfo = ParaseleneWildRace.WindowInfo:new(o.optionsWindowInfo)
    self.trackMasterOptions = ParaseleneWildRace.TrackMasterOptions:new(o.trackMasterOptions)
    self.uiOptions = ParaseleneWildRace.UIOptions:new(o.uiOptions)
    self.worldMarkerOptions = ParaseleneWildRace.WorldMarkerOptions:new(o.worldMarkerOptions)

    local clientWindowInfo = self.CreateDefaultClientWindowInfo()
    self.shareWindowInfo = self.CreateDefaultShareWindowInfo(clientWindowInfo)
    self.areWaypointsVisible = self.CreateDefaultAreWaypointsVisible()
    self.currentCourse = self.CreateDefaultCurrentCourse()

    return self:IsValid()
end

-- Version 5 was only ever distributed to a few people to test the Share window and communications.
function ParaseleneWildRace.CharacterData:RestoreV5(o)
    -- self.version is not restored here.
    self.clientWindowInfo = ParaseleneWildRace.WindowInfo:new(o.clientWindowInfo)
    self.editorWindowInfo = ParaseleneWildRace.WindowInfo:new(o.editorWindowInfo)
    self.filesWindowInfo = ParaseleneWildRace.WindowInfo:new(o.filesWindowInfo)
    self.optionsWindowInfo = ParaseleneWildRace.WindowInfo:new(o.optionsWindowInfo)
    self.shareWindowInfo = ParaseleneWildRace.WindowInfo:new(o.shareWindowInfo)
    self.trackMasterOptions = ParaseleneWildRace.TrackMasterOptions:new(o.trackMasterOptions)
    self.uiOptions = ParaseleneWildRace.UIOptions:new(o.uiOptions)
    self.worldMarkerOptions = ParaseleneWildRace.WorldMarkerOptions:new(o.worldMarkerOptions)

    self.areWaypointsVisible = self.CreateDefaultAreWaypointsVisible()
    self.currentCourse = self.CreateDefaultCurrentCourse()

    return self:IsValid()
end

function ParaseleneWildRace.CharacterData:RestoreV6(o)
    -- self.version is not restored here.
    self.clientWindowInfo = ParaseleneWildRace.WindowInfo:new(o.clientWindowInfo)
    self.editorWindowInfo = ParaseleneWildRace.WindowInfo:new(o.editorWindowInfo)
    self.filesWindowInfo = ParaseleneWildRace.WindowInfo:new(o.filesWindowInfo)
    self.optionsWindowInfo = ParaseleneWildRace.WindowInfo:new(o.optionsWindowInfo)
    self.shareWindowInfo = ParaseleneWildRace.WindowInfo:new(o.shareWindowInfo)
    self.trackMasterOptions = ParaseleneWildRace.TrackMasterOptions:new(o.trackMasterOptions)
    self.uiOptions = ParaseleneWildRace.UIOptions:new(o.uiOptions)
    self.worldMarkerOptions = ParaseleneWildRace.WorldMarkerOptions:new(o.worldMarkerOptions)

    if type(o.areWaypointsVisible) == "boolean" then
        self.areWaypointsVisible = o.areWaypointsVisible
    end
    self.currentCourse = ParaseleneWildRace.Course:new(o.currentCourse)

    return self:IsValid()
end

function ParaseleneWildRace.CharacterData:Save()
    return {
        version = CHARACTER_DATA_FORMAT_VERSION,
        clientWindowInfo = self.clientWindowInfo:Save(),
        editorWindowInfo = self.editorWindowInfo:Save(),
        filesWindowInfo = self.filesWindowInfo:Save(),
        optionsWindowInfo = self.optionsWindowInfo:Save(),
        shareWindowInfo = self.shareWindowInfo:Save(),
        trackMasterOptions = self.trackMasterOptions:Save(),
        uiOptions = self.uiOptions:Save(),
        worldMarkerOptions = self.worldMarkerOptions:Save(),
        areWaypointsVisible = self.areWaypointsVisible,
        currentCourse = self.currentCourse:Save()
    }
end
