-- Paraselene's WildRace
-- Version 0.5.2

require "Apollo"
require "GameLib"
require "ICCommLib"
require "ICComm"
require "Window"
 
local LOAD_TIMER_TICK_INTERVAL_SECS = 1.0
local LOAD_TIMER_EXPIRE_SECS = 15.0
local UPDATE_TIMER_TICK_INTERVAL_SECS = 1.0

local STAGE_UNTRIGGERED = 0 -- The action has not happened yet.
local STAGE_TRIGGERED = 1   -- The action has happened, but no message has been posted.
local STAGE_REPORTED = 2    -- The action has happened and a message has been posted.

function ParaseleneWildRace:new()
    local i = {}
    setmetatable(i, self)
    self.__index = self

    i.interfaceMenuListLoadedStage = STAGE_UNTRIGGERED
    i.docLoadedStage = STAGE_UNTRIGGERED
    i.realmCharacterInfoLoadedStage = STAGE_UNTRIGGERED
    i.commChannelOpenedStage = STAGE_UNTRIGGERED
    i.accountDataLoadedStage = STAGE_UNTRIGGERED
    i.characterDataLoadedStage = STAGE_UNTRIGGERED

    return i
end

function ParaseleneWildRace:Init()
    math.randomseed(os.time())
    math.random() -- Some random number generators deliver the same first random numer.
    math.random() -- Pump a few random numbers out for the randomness to get going.
    math.random()

    local hasConfigureFunction = true
    local configureButtonTitle = ParaseleneWildRace.appNameFull
    local dependencies = {}
    Apollo.RegisterAddon(self, hasConfigureFunction, configureButtonTitle, dependencies)
end

function ParaseleneWildRace:Initialize()
    if self.accountDataLoadedStage == STAGE_UNTRIGGERED then
        ChatSystemLib.PostOnChannel(
            ChatSystemLib.ChatChannel_System,
            ParaseleneWildRace.appName .. ": Account-level data not loaded."
        )
    else
        self.accountData = ParaseleneWildRace.AccountData:new(self.loadedAccountData)
        if not self.accountData then
            ChatSystemLib.PostOnChannel(
                ChatSystemLib.ChatChannel_System,
                ParaseleneWildRace.appName .. ": Account-level data is corrupt. Not using it."
            )
        end
    end
    if not self.accountData then
        self.accountData = ParaseleneWildRace.AccountData.CreateDefault()
        ChatSystemLib.PostOnChannel(
            ChatSystemLib.ChatChannel_System,
            ParaseleneWildRace.appName .. ": Account-level data set to default."
        )
    end

    if self.characterDataLoadedStage == STAGE_UNTRIGGERED then
        ChatSystemLib.PostOnChannel(
            ChatSystemLib.ChatChannel_System,
            ParaseleneWildRace.appName .. ": Character-level data not loaded."
        )
    else
        self.characterData = ParaseleneWildRace.CharacterData:new(self.loadedCharacterData)
        if not self.characterData then
            ChatSystemLib.PostOnChannel(
                ChatSystemLib.ChatChannel_System,
                ParaseleneWildRace.appName .. ": Character-level data is corrupt. Not using it."
            )
        end
    end
    if not self.characterData then
        self.characterData = ParaseleneWildRace.CharacterData.CreateDefault()
        ChatSystemLib.PostOnChannel(
            ChatSystemLib.ChatChannel_System,
            ParaseleneWildRace.appName .. ": Character-level data set to default."
        )
    end

    self.about = ParaseleneWildRace.About:new(self)
    if not self.about then
        return
    end

    self.client = ParaseleneWildRace.Client:new(self)
    if not self.client then
        return
    end

    self.editor = ParaseleneWildRace.Editor:new(self)
    if not self.editor then
        return
    end

    self.files = ParaseleneWildRace.Files:new(self)
    if not self.files then
        return
    end

    self.options = ParaseleneWildRace.Options:new(self)
    if not self.options then
        return
    end

    self.share = ParaseleneWildRace.Share:new(self)
    if not self.share then
        return
    end

    self.trackMaster = ParaseleneWildRace.TrackMaster:new(self)

    self.worldMarker = ParaseleneWildRace.WorldMarker:new(self)
    if not self.worldMarker then
        return
    end

    self.client:SetCourse(self.characterData.currentCourse)
    self.files:SetStorage(self.accountData.storage)
    self.options:AdjustOptions()

    Event_FireGenericEvent(
        "OneVersion_ReportAddonInfo",
        ParaseleneWildRace.appNameFull,
        ParaseleneWildRace.appVersion.major,
        ParaseleneWildRace.appVersion.minor,
        ParaseleneWildRace.appVersion.patch)

    self.updateTimer = ApolloTimer.Create(UPDATE_TIMER_TICK_INTERVAL_SECS, true, "OnUpdateTimer", self)

    Apollo.RegisterSlashCommand("wildrace", "OnSlashCommand", self)

    ChatSystemLib.PostOnChannel(
        ChatSystemLib.ChatChannel_System,
        ParaseleneWildRace.appName .. ": Initialization succeeded."
    )

    self.isInitialized = true
end

function ParaseleneWildRace:LoadRealmCharacterInfo()
    local me = GameLib.GetPlayerUnit()

    if not self.realm then
        self.realm = GameLib.GetRealmName()
        self.realm = self.realm and string.gsub(self.realm, "^%s*(.-)%s*$", "%1") -- Strip leading and trailing spaces.
    end

    if not self.name then
        self.name = me and me:GetName()
        self.name = self.name and string.gsub(self.name, "^%s*(.-)%s*$", "%1") -- Strip leading and trailing spaces.
    end

    if not self.faction then
        local faction = me and me:GetFaction()
        if faction == Unit.CodeEnumFaction.ExilesPlayer then
            self.faction = ParaseleneWildRace.waypointFaction.exile
        elseif faction == Unit.CodeEnumFaction.DominionPlayer then
            self.faction = ParaseleneWildRace.waypointFaction.dominion
        end
    end

    return (self.realm ~= nil) and
           (self.name ~= nil) and
           (self.faction ~= nil)
end

function ParaseleneWildRace:OnCommMessageReceived(channel, strMessage, strSender)
    local start = 1
    local prefix = string.sub(strMessage, start, start + ParaseleneWildRace.commPrefixLength - 1)
    if prefix == ParaseleneWildRace.commPrefix then
        start = start + ParaseleneWildRace.commPrefixLength
        local version = string.sub(strMessage, start, start + ParaseleneWildRace.commVersionLength - 1)
        if version == ParaseleneWildRace.commVersion then
            start = start + ParaseleneWildRace.commVersionLength
            local packetType = string.sub(strMessage, start, start + ParaseleneWildRace.commTypeLength - 1)
            if self.share and (packetType == ParaseleneWildRace.commType.share) then
                start = start + ParaseleneWildRace.commTypeLength
                self.share:OnCommMessageReceived(strSender, string.sub(strMessage, start))
            end
        end
    end
end

function ParaseleneWildRace:OnConfigure()
    self:ToggleWndClientVisibility()
end

function ParaseleneWildRace:OnDocLoaded()
    ChatSystemLib.PostOnChannel(
        ChatSystemLib.ChatChannel_System,
        ParaseleneWildRace.appNameFull .. ": Version " .. "0.5.2" .. " initializing."
    )
    ChatSystemLib.PostOnChannel(
        ChatSystemLib.ChatChannel_System,
        ParaseleneWildRace.appName .. ": Initialization may take up to " ..
        string.format("%d", LOAD_TIMER_EXPIRE_SECS) .. " seconds."
    )
    self.loadStartClock = os.clock()
    self.loadTimer = ApolloTimer.Create(LOAD_TIMER_TICK_INTERVAL_SECS, true, "OnLoadTimer", self)
end

function ParaseleneWildRace:OnInterfaceMenuListHasLoaded()
    Apollo.RegisterEventHandler("InterfaceMenuToggleWildRace", "ToggleWndClientVisibility", self)
    Event_FireGenericEvent(
        "InterfaceMenuList_NewAddOn",
        ParaseleneWildRace.appNameFull,
        {"InterfaceMenuToggleWildRace", "", ParaseleneWildRace.appIcon}
    )
    self.interfaceMenuListLoadedStage = STAGE_TRIGGERED
end

function ParaseleneWildRace:OnLoad()
    Apollo.RegisterEventHandler("InterfaceMenuListHasLoaded", "OnInterfaceMenuListHasLoaded", self)
    self.doc = XmlDoc.CreateFromFile("ParaseleneWildRace.xml")
    self.doc:RegisterCallback("OnDocLoaded", self)
end

function ParaseleneWildRace:OnLoadTimer()
    local currentClock = os.clock()
    local elapsed = (((currentClock - self.loadStartClock) >= 0) and (currentClock - self.loadStartClock)) or 0

    if self.interfaceMenuListLoadedStage == STAGE_TRIGGERED then
        self.interfaceMenuListLoadedStage = STAGE_REPORTED
        --[[
        ChatSystemLib.PostOnChannel(
            ChatSystemLib.ChatChannel_System,
            ParaseleneWildRace.appName .. ": Interface menu item installed."
        )
        --]]
    end

    if (self.docLoadedStage == STAGE_UNTRIGGERED) and self.doc:IsLoaded() then
        self.docLoadedStage = STAGE_REPORTED
        --[[
        ChatSystemLib.PostOnChannel(
            ChatSystemLib.ChatChannel_System,
            ParaseleneWildRace.appName .. ": Interface document loaded."
        )
        --]]
    end

    if (self.realmCharacterInfoLoadedStage == STAGE_UNTRIGGERED) and self:LoadRealmCharacterInfo() then
        self.realmCharacterInfoLoadedStage = STAGE_REPORTED
        --[[
        ChatSystemLib.PostOnChannel(
            ChatSystemLib.ChatChannel_System,
            ParaseleneWildRace.appName .. ": Realm/Character information loaded."
        )
        --]]
    end

    if (self.commChannelOpenedStage == STAGE_UNTRIGGERED) and self:OpenComm() then
        self.commChannelOpenedStage = STAGE_REPORTED
        --[[
        ChatSystemLib.PostOnChannel(
            ChatSystemLib.ChatChannel_System,
            ParaseleneWildRace.appName .. ": Communications channel opened."
        )
        --]]
    end

    if self.accountDataLoadedStage == STAGE_TRIGGERED then
        self.accountDataLoadedStage = STAGE_REPORTED
        --[[
        ChatSystemLib.PostOnChannel(
            ChatSystemLib.ChatChannel_System,
            ParaseleneWildRace.appName .. ": Account-level data loaded."
        )
        --]]
    end

    if self.characterDataLoadedStage == STAGE_TRIGGERED then
        self.characterDataLoadedStage = STAGE_REPORTED
        --[[
        ChatSystemLib.PostOnChannel(
            ChatSystemLib.ChatChannel_System,
            ParaseleneWildRace.appName .. ": Character-level data loaded."
        )
        --]]
    end

    -- The following condition must become true before the timeout, otherwise we error out.
    if (self.docLoadedStage >= STAGE_TRIGGERED) and
       (self.interfaceMenuListLoadedStage >= STAGE_TRIGGERED) and
       (self.realmCharacterInfoLoadedStage >= STAGE_TRIGGERED) and
       (self.commChannelOpenedStage >= STAGE_TRIGGERED) then

        -- Initialize once the timer has expired, or the restore operations are complete.
        if (elapsed >= LOAD_TIMER_EXPIRE_SECS) or
            ((self.accountDataLoadedStage >= STAGE_TRIGGERED) and
             (self.characterDataLoadedStage >= STAGE_TRIGGERED)) then
            self.loadTimer:Stop()
            self.loadTimer = nil
            self:Initialize()
            return
        end
    end

    -- Check for timeout.
    if elapsed >= LOAD_TIMER_EXPIRE_SECS then
        self.loadTimer:Stop()
        self.loadTimer = nil
        ChatSystemLib.PostOnChannel(
            ChatSystemLib.ChatChannel_System,
            ParaseleneWildRace.appName .. ": Error: Initialization failed. It took too much time."
        )
        Apollo.AddAddonErrorText(self, message)
    end
end

function ParaseleneWildRace:OnRestore(eLevel, tData)
    self.loadStartClock = os.clock() -- Reset the load timer to allow for lengthy restores.
    if eLevel == GameLib.CodeEnumAddonSaveLevel.Account then
        self.loadedAccountData = tData
        self.accountDataLoadedStage = STAGE_TRIGGERED
    elseif eLevel == GameLib.CodeEnumAddonSaveLevel.Character then
        self.loadedCharacterData = tData
        self.characterDataLoadedStage = STAGE_TRIGGERED
    end
end

function ParaseleneWildRace:OnSave(eLevel)
    -- Could not think of another place to detect a reload or exit.
    if self.updateTimer then
        self.updateTimer:Stop()
        self.updateTimer = nil
    end

    if (eLevel == GameLib.CodeEnumAddonSaveLevel.Account) and self.accountData then
        return self:SaveAccountData()
    elseif (eLevel == GameLib.CodeEnumAddonSaveLevel.Character) and self.characterData then
        return self:SaveCharacterData()
    end
end

function ParaseleneWildRace:OnSlashCommand()
    self:ToggleWndClientVisibility()
end

function ParaseleneWildRace:OpenComm()
    if not self.comm then
        self.comm = ICCommLib.JoinChannel(ParaseleneWildRace.commChannel, ICCommLib.CodeEnumICCommChannelType.Global)
        if self.comm then
            self.comm:SetReceivedMessageFunction("OnCommMessageReceived", self)
        end
    end
    return (self.comm ~= nil)
end

function ParaseleneWildRace:OnUpdateTimer()
    self.client:OnUpdateTimer()
    self.share:OnUpdateTimer()
end

function ParaseleneWildRace:SaveAccountData()
    self.accountData.storage = self.files and self.files.storage
    return self.accountData:Save()
end

function ParaseleneWildRace:SaveCharacterData()
    self.characterData.currentCourse = self.client and self.client.course
    return self.characterData:Save()
end

function ParaseleneWildRace:SendMessage(message)
    self.comm:SendMessage(ParaseleneWildRace.commPrefix .. ParaseleneWildRace.commVersion .. message)
end

function ParaseleneWildRace:SendPrivateMessage(recipient, message)
    self.comm:SendPrivateMessage(recipient, ParaseleneWildRace.commPrefix .. ParaseleneWildRace.commVersion .. message)
end

function ParaseleneWildRace:ToggleWndClientVisibility()
    if self.isInitialized then
        self.client:ToggleWindowVisibility()
        Sound.Play(ParaseleneWildRace.buttonClickSound)
    end
end

ParaseleneWildRaceInstance = ParaseleneWildRace:new()
ParaseleneWildRaceInstance:Init()
