-- Paraselene's WildRace
-- Version 0.5.2

-- Maximum length of a received string.
local MAX_STRING_LENGTH = 999
local STRING_LENGTH_PRECISION = 3 -- Enough digits to represent MAX_STRING_LENGTH.

-- Minimum and maximum value of a received coordinate.
local MIN_COORDINATE = -99999
local MAX_COORDINATE = 99999
local COORDINATE_PRECISION = 5 -- Enough digits to represent MIN_COORDINATE and MAX_COORDINATE.

-- Maximum number of waypoints in a received course.
local MAX_WAYPOINTS_PER_COURSE = 99
local MAX_WAYPOINTS_PER_COURSE_PRECISION = 2 -- Enough digits to represent MAX_WAYPOINTS_PER_COURSE.

ParaseleneWildRace.Deserialize = {}
ParaseleneWildRace.Serialize = {}

function ParaseleneWildRace.Deserialize.Boolean(serial)
    if serial then
        local value, remaining = string.match(serial, "^([+-])(.*)")
        if value then
            return (value == "+"), remaining
        end
    end
    return nil
end

function ParaseleneWildRace.Deserialize.CourseInfo(serial)
    if serial then
        local name, waypointCount, strictMode, timedMode, timeout, remaining

        name, remaining = ParaseleneWildRace.Deserialize.String(serial)
        _, remaining = ParaseleneWildRace.Deserialize.Literal(remaining, ",")
        waypointCount, remaining = ParaseleneWildRace.Deserialize.Number{
            serial = remaining,
            highLimit = MAX_WAYPOINTS_PER_COURSE,
            lowLimit = 0,
            negativeAllowed = false,
            precision = MAX_WAYPOINTS_PER_COURSE_PRECISION
        }
        _, remaining = ParaseleneWildRace.Deserialize.Literal(remaining, ",")
        strictMode, remaining = ParaseleneWildRace.Deserialize.Boolean(remaining)
        _, remaining = ParaseleneWildRace.Deserialize.Literal(remaining, ",")
        timedMode, remaining = ParaseleneWildRace.Deserialize.Boolean(remaining)
        _, remaining = ParaseleneWildRace.Deserialize.Literal(remaining, ",")
        timeout, remaining = ParaseleneWildRace.Deserialize.Number{
            serial = remaining,
            highLimit = ParaseleneWildRace.maxRaceTime,
            lowLimit = 0,
            negativeAllowed = false,
            precision = ParaseleneWildRace.maxRaceTimeDigits
        }
        if name and waypointCount and (strictMode ~= nil) and (timedMode ~= nil) and
           timeout and (timeout <= ParaseleneWildRace.maxRaceTime) then
            return {
                name = name,
                waypointCount = waypointCount,
                strictMode = strictMode,
                timedMode = timedMode,
                timeout = timeout
            }, remaining
        end
    end
    return nil
end

function ParaseleneWildRace.Deserialize.Faction(serial)
    if serial then
        local value, remaining = string.match(serial, "^([ade])(.*)")
        if value then
            local faction = ((value == "a") and ParaseleneWildRace.waypointFaction.all) or
                            ((value == "d") and ParaseleneWildRace.waypointFaction.dominion) or
                            ((value == "e") and ParaseleneWildRace.waypointFaction.exile)
            return faction, remaining
        end
    end
    return nil
end

function ParaseleneWildRace.Deserialize.Literal(serial, literal)
    local length = string.len(literal)
    if serial and (length <= string.len(serial)) and (string.sub(serial, 1, length) == literal) then
        return literal, string.sub(serial, length + 1)
    end
    return nil
end

function ParaseleneWildRace.Deserialize.Number(args)
    if args.serial then
        local pattern = (args.negativeAllowed and "^([-]?%d+)(.*)") or "^(%d+)(.*)"
        local value, remaining = string.match(args.serial, pattern)
        if value and (string.len(value) <= (args.precision + 1)) then
            local number = tonumber(value)
            local maxLength = ((number < 0) and (args.precision + 1)) or args.precision
            if (string.len(value) <= maxLength) and
               (number >= args.lowLimit) and (number <= args.highLimit) then
                return number, remaining
            end
        end
    end
    return nil
end

function ParaseleneWildRace.Deserialize.Point(serial)
    if serial then
        local function DeserializeCoordinate(serial)
            return ParaseleneWildRace.Deserialize.Number{
                serial = serial,
                highLimit = MAX_COORDINATE,
                lowLimit = MIN_COORDINATE,
                negativeAllowed = true,
                precision = COORDINATE_PRECISION
            }
        end

        local x, y, z, remaining

        x, remaining = DeserializeCoordinate(serial)
        _, remaining = ParaseleneWildRace.Deserialize.Literal(remaining, ",")
        y, remaining = DeserializeCoordinate(remaining)
        _, remaining = ParaseleneWildRace.Deserialize.Literal(remaining, ",")
        z, remaining = DeserializeCoordinate(remaining)
        return ParaseleneWildRace.Point:new{
            x = tonumber(x),
            y = tonumber(y),
            z = tonumber(z)
        }, remaining
    end
    return nil
end

function ParaseleneWildRace.Deserialize.String(serial)
    if serial then
        local length, remaining = ParaseleneWildRace.Deserialize.Number{
            serial = serial,
            highLimit = MAX_STRING_LENGTH,
            lowLimit = 0,
            negativeAllowed = false,
            precision = STRING_LENGTH_PRECISION
        }
        _, remaining = ParaseleneWildRace.Deserialize.Literal(remaining, ",")
        if remaining then
            length = tonumber(length)
            if length <= string.len(remaining) then
                return string.sub(remaining, 1, length), string.sub(remaining, length + 1)
            end
        end
    end
    return nil
end

function ParaseleneWildRace.Deserialize.Waypoint(serial)
    if serial then
        local name, realm, zone, subZone, inHousing, faction, point, remaining

        name, remaining = ParaseleneWildRace.Deserialize.String(serial)
        _, remaining = ParaseleneWildRace.Deserialize.Literal(remaining, ",")
        realm, remaining = ParaseleneWildRace.Deserialize.String(remaining)
        _, remaining = ParaseleneWildRace.Deserialize.Literal(remaining, ",")
        zone, remaining = ParaseleneWildRace.Deserialize.String(remaining)
        _, remaining = ParaseleneWildRace.Deserialize.Literal(remaining, ",")
        subZone, remaining = ParaseleneWildRace.Deserialize.String(remaining)
        _, remaining = ParaseleneWildRace.Deserialize.Literal(remaining, ",")
        inHousing, remaining = ParaseleneWildRace.Deserialize.Boolean(remaining)
        _, remaining = ParaseleneWildRace.Deserialize.Literal(remaining, ",")
        faction, remaining = ParaseleneWildRace.Deserialize.Faction(remaining)
        _, remaining = ParaseleneWildRace.Deserialize.Literal(remaining, ",")
        point, remaining = ParaseleneWildRace.Deserialize.Point(remaining)
        return ParaseleneWildRace.Waypoint:new{
            name = name,
            realm = realm,
            zone = zone,
            subZone = subZone,
            inHousing = inHousing,
            faction = faction,
            point = point
        }, remaining
    end
    return nil
end

function ParaseleneWildRace.Serialize.Boolean(bool)
    return (bool and "+") or "-"
end

function ParaseleneWildRace.Serialize.CourseInfo(course)
    return ParaseleneWildRace.Serialize.String(course.name) .. "," ..
           ParaseleneWildRace.Serialize.Number(#course.waypoints) .. "," ..
           ParaseleneWildRace.Serialize.Boolean(course.strictMode) .. "," ..
           ParaseleneWildRace.Serialize.Boolean(course.timedMode) .. "," ..
           ParaseleneWildRace.Serialize.Number(course.timeout)
end

function ParaseleneWildRace.Serialize.CRC16(serial)
    local crc = 0xffff
    for i = 1, #serial do
        local c = serial:byte(i)
        crc = bit32.bxor(crc, c)
        for j = 1, 8 do
            local k = bit32.band(crc, 1)
            crc = bit32.rshift(crc, 1)
            if k ~= 0 then
                crc = bit32.bxor(crc, 0x8408)
            end
        end
    end
    return crc
end

function ParaseleneWildRace.Serialize.Faction(faction)
    return (((faction == ParaseleneWildRace.waypointFaction.all) and "a") or
            ((faction == ParaseleneWildRace.waypointFaction.dominion) and "d") or
            ((faction == ParaseleneWildRace.waypointFaction.exile) and "e"))
end

function ParaseleneWildRace.Serialize.Number(num)
    return string.format("%.0f", num)
end

function ParaseleneWildRace.Serialize.Point(point)
    return ParaseleneWildRace.Serialize.Number(point.x) .. "," ..
           ParaseleneWildRace.Serialize.Number(point.y) .. "," ..
           ParaseleneWildRace.Serialize.Number(point.z)
end

function ParaseleneWildRace.Serialize.String(str)
    return ParaseleneWildRace.Serialize.Number(string.len(str)) .. "," .. str
end

function ParaseleneWildRace.Serialize.Waypoint(waypoint)
    return ParaseleneWildRace.Serialize.String(waypoint.name) .. "," ..
           ParaseleneWildRace.Serialize.String(waypoint.realm) .. "," ..
           ParaseleneWildRace.Serialize.String(waypoint.zone) .. "," ..
           ParaseleneWildRace.Serialize.String(waypoint.subZone) .. "," ..
           ParaseleneWildRace.Serialize.Boolean(waypoint.inHousing) .. "," ..
           ParaseleneWildRace.Serialize.Faction(waypoint.faction) .. "," ..
           ParaseleneWildRace.Serialize.Point(waypoint.point)
end
